#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "config.h"
#include "lib.h"
#include "framebuf.h"

static int screen_mode   = 0;
static int screen_width  = 0;
static int screen_height = 0;
static int screen_flush  = 0;
static int screen_cursol = 0;
static int cursol_x = 0;
static int cursol_y = 0;
static int cursol_attr = 0;

static struct linebuf {
  unsigned char charbuf[FRAMEBUF_MAXWIDTH];
  unsigned int attrbuf[FRAMEBUF_MAXWIDTH];
} linebuf[FRAMEBUF_MAXHEIGHT];

static struct {
  unsigned int flags;
#define BUFFER_FLAG_DIRTY (1 <<  0)
  struct linebuf *linebuf;
} framebuf[FRAMEBUF_MAXHEIGHT];

#define DEFAULT_CHAR (' ')
#define DEFAULT_ATTR ((0 << 0) | (9 << 4) | (9 << 8))

int framebuf_get_screen_mode(void)
{
  return screen_mode;
}

int framebuf_get_screen_width(void)
{
  return screen_width;
}

int framebuf_get_screen_height(void)
{
  return screen_height;
}

int framebuf_get_cursol_x(void)
{
  return cursol_x;
}

int framebuf_get_cursol_y(void)
{
  return cursol_y;
}

int framebuf_init(void)
{
  return framebuf_setscreen(0, 0, 0, 0, 0);
}

static int _flush(void)
{
  int i, n, size;
  unsigned char *cp;
  unsigned int attr = -1, *ap;

  if (screen_mode == 0)
    return 0;

  printf("\x1b[1;1H"); /* Home */

  for (i = 0; i < screen_height; i++) {
    if (framebuf[i].flags & BUFFER_FLAG_DIRTY) {
      n = 0;
      cp = framebuf[i].linebuf->charbuf;
      ap = framebuf[i].linebuf->attrbuf;
      while (n < screen_width) {
	if (attr != *ap) {
	  attr = *ap;
	  printf("\x1b[%dm",  0 + ((attr >> 0) & 0xf));
	  printf("\x1b[%dm", 30 + ((attr >> 4) & 0xf));
	  printf("\x1b[%dm", 40 + ((attr >> 8) & 0xf));
	}
	for (size = 0; n + size < screen_width; size++, ap++) {
	  if (*ap != attr)
	    break;
	}
	fwrite(cp, size, 1, stdout);
	cp += size;
	n  += size;
      }
      framebuf[i].flags &= ~BUFFER_FLAG_DIRTY;
    }
    printf("\x1b[1E");
  }
  fflush(stdout);

  return 0;
}

static int linebuf_clear(struct linebuf *linebuf, int ch, unsigned int attr)
{
  int i;
  for (i = 0; i < screen_width; i++) {
    linebuf->charbuf[i] = ch;
    linebuf->attrbuf[i] = attr;
  }
  return 0;
}

int framebuf_setscreen(int mode, int width, int height, int flush, int cursol)
{
  int i;

  mode = (mode < 0) ? screen_mode : mode;

  switch (mode) {
  case 0:
    break;

  case 1:
    width  = (width  < 0) ? screen_width  : width;
    height = (height < 0) ? screen_height : height;
    flush  = (flush  < 0) ? screen_flush  : flush;
    cursol = (cursol < 0) ? screen_cursol : cursol;

    width  = ((width  > 0) && (width  <= FRAMEBUF_MAXWIDTH ))
      ? width  : FRAMEBUF_DEFAULT_WIDTH;
    height = ((height > 0) && (height <= FRAMEBUF_MAXHEIGHT))
      ? height : FRAMEBUF_DEFAULT_HEIGHT;

    if ((width != screen_width) || (height != screen_height)) {
      screen_width  = width;
      screen_height = height;
      cursol_x = 0;
      cursol_y = 0;
      cursol_attr = DEFAULT_ATTR;

      for (i = 0; i < screen_height; i++) {
	framebuf[i].linebuf = &linebuf[i];
	linebuf_clear(framebuf[i].linebuf, DEFAULT_CHAR, DEFAULT_ATTR);
	framebuf[i].flags |= BUFFER_FLAG_DIRTY;
      }
    }

    screen_flush  = flush;
    screen_cursol = cursol;

    break;

  default:
    return NLL_ERRCODE_FORMULA_INVALID_PARAMETER;
  }

  screen_mode = mode;

  if (screen_flush) {
    for (i = 0; i < screen_height; i++)
      framebuf[i].flags |= BUFFER_FLAG_DIRTY;
    _flush();
  }

  return 0;
}

int framebuf_clear(int ch, int flush)
{
  int i;

  if (screen_mode == 0) {
    if (ch >= 0) {
      printf("\x1b[1;1H"); /* Home */
      printf("\x1b[2J"); /* Clear */
      fflush(stdout);
    }
    return 0;
  }

  if (ch >= 0) {
    ch = ch ? ch : DEFAULT_CHAR;
    cursol_x = 0;
    cursol_y = 0;
    for (i = 0; i < screen_height; i++) {
      linebuf_clear(framebuf[i].linebuf, ch, DEFAULT_ATTR);
      framebuf[i].flags |= BUFFER_FLAG_DIRTY;
    }
  }

  if (screen_flush || (flush >= 0)) {
    for (i = 0; i < screen_height; i++)
      framebuf[i].flags |= BUFFER_FLAG_DIRTY;
    _flush();
  }

  return 0;
}

int framebuf_reset(void)
{
  if (screen_mode == 0) {
    printf("\x1b[39m"); /* Reset foreground color */
    printf("\x1b[49m"); /* Reset background color */
    printf("\x1b[0m"); /* Reset attr */
    fflush(stdout);
    return 0;
  }

  cursol_attr = DEFAULT_ATTR;

  return 0;
}

int framebuf_setcolor(int color)
{
  if (screen_mode == 0) {
    /*
     *  0- 7: Foreground color
     *     9: Reset foreground color
     * 10-17: Background color
     *    19: Reset background color
     */
    if (color < 0) {
      printf("\x1b[39m"); /* Reset foreground color */
      printf("\x1b[49m"); /* Reset background color */
    } else {
      printf("\x1b[%dm", 30 + color);
    }
    fflush(stdout);
    return 0;
  }

  if (color < 0) {
    /* none */
  } else if (color < 10) {
    cursol_attr &= ~(0xf << 4);
    cursol_attr |= color << 4;
  } else if (color < 20) {
    color -= 10;
    cursol_attr &= ~(0xf << 8);
    cursol_attr |= color << 8;
  }

  return 0;
}

int framebuf_setattr(int attr)
{
  if (screen_mode == 0) {
    printf("\x1b[%dm", attr); /* Mode (0:Reset) */
    fflush(stdout);
    return 0;
  }

  if (attr < 0) {
    /* none */
  } else if (attr < 16) {
    cursol_attr &= ~(0xf << 0);
    cursol_attr |= attr << 0;
  }

  return 0;
}

int framebuf_move_cursol(int x, int y)
{
  x = (x < 0) ? cursol_x : x;
  y = (y < 0) ? cursol_y : y;

  if (screen_mode == 0) {
    printf("\x1b[%d;%dH", y + 1, x + 1);
    fflush(stdout);
    return 0;
  }

  cursol_x = x;
  cursol_y = y;

  return 0;
}

int framebuf_move_up(int num)
{
  if (screen_mode == 0) {
    printf("\x1b[%dA", num); /* Up */
    fflush(stdout);
    return 0;
  }
  if (cursol_y > 0)
    cursol_y -= num;
  return 0;
}

int framebuf_move_down(int num)
{
  if (screen_mode == 0) {
    printf("\x1b[%dB", num); /* Down */
    fflush(stdout);
    return 0;
  }
  if (cursol_y < screen_height - 1)
    cursol_y += num;
  return 0;
}

int framebuf_move_left(int num)
{
  if (screen_mode == 0) {
    printf("\x1b[%dD", num); /* Left */
    fflush(stdout);
    return 0;
  }
  if (cursol_x > 0)
    cursol_x -= num;
  return 0;
}

int framebuf_move_right(int num)
{
  if (screen_mode == 0) {
    printf("\x1b[%dC", num); /* Right */
    fflush(stdout);
    return 0;
  }
  if (cursol_x < screen_width - 1)
    cursol_x += num;
  return 0;
}

int framebuf_scroll_up(int num)
{
  int i;
  struct linebuf *buf;

  if (screen_mode == 0) {
    printf("\x1b[%dT", num); /* Scroll Up */
    fflush(stdout);
    return 0;
  }

  for (i = screen_height - 1; i >= 0; i--) {
    if (i >= num) {
      buf = framebuf[i].linebuf;
      framebuf[i].linebuf = framebuf[i - num].linebuf;
      framebuf[i - num].linebuf = buf;
    } else {
      linebuf_clear(framebuf[i].linebuf, DEFAULT_CHAR, DEFAULT_ATTR);
    }
    framebuf[i].flags |= BUFFER_FLAG_DIRTY;
  }

  if (screen_flush)
    _flush();

  return 0;
}

int framebuf_scroll_down(int num)
{
  int i;
  struct linebuf *buf;

  if (screen_mode == 0) {
    printf("\x1b[%dS", num); /* Scroll Down */
    fflush(stdout);
    return 0;
  }

  for (i = 0; i < screen_height; i++) {
    if (i + num < screen_height) {
      buf = framebuf[i].linebuf;
      framebuf[i].linebuf = framebuf[i + num].linebuf;
      framebuf[i + num].linebuf = buf;
    } else {
      linebuf_clear(framebuf[i].linebuf, DEFAULT_CHAR, DEFAULT_ATTR);
    }
    framebuf[i].flags |= BUFFER_FLAG_DIRTY;
  }

  if (screen_flush)
    _flush();

  return 0;
}

int framebuf_cursol_on(void)
{
  printf("\x1b[>5l"); /* Cursol ON */
  fflush(stdout);
  return 0;
}

int framebuf_cursol_off(void)
{
  printf("\x1b[>5h"); /* Cursol OFF */
  fflush(stdout);
  return 0;
}

int framebuf_puts(const char *str)
{
  int len, i;

  if (screen_mode == 0) {
    fputs(str, stdout);
    fflush(stdout);
    return 0;
  }

  len = strlen(str);

  for (i = 0; i < len; i++) {
    switch (str[i]) {
    case '\n':
      goto newline;

    default:
      if ((cursol_x < screen_width) && (cursol_y < screen_height)) {
	framebuf[cursol_y].linebuf->charbuf[cursol_x] = str[i];
	framebuf[cursol_y].linebuf->attrbuf[cursol_x] = cursol_attr;
	framebuf[cursol_y].flags |= BUFFER_FLAG_DIRTY;
	if (++cursol_x >= screen_width)
	  goto newline;
      }
      break;

    newline:
      cursol_x = 0;
      if (cursol_y < screen_height - 1)
	cursol_y++;
      else
	framebuf_scroll_down(1);
      break;
    }
  }

  if (screen_flush)
    _flush();

  return 0;
}

int framebuf_getch(int x, int y)
{
  x = (x < 0) ? cursol_x : x;
  y = (y < 0) ? cursol_y : y;
  return framebuf[y].linebuf->charbuf[x];
}
