#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#ifdef USE_MATHLIB
#include <math.h>
#endif
#ifndef WIN32
#include <sys/socket.h>
#else
#include <winsock2.h>
#endif

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "config.h"
#include "types.h"
#include "lib.h"
#include "value.h"
#include "string.h"
#include "variable.h"

static struct const_variable {
  char *name;
  value_type_t type;
  integer_t integer;
  char *string;
  void *pointer;
} const_variables[] = {
  { "NULL"        , VALUE_TYPE_POINTER  , 0           , NULL, NULL },
#ifdef O_RDONLY
  { "O_RDONLY"    , VALUE_TYPE_INTEGER  , O_RDONLY    , NULL, NULL },
#endif
#ifdef O_WRONLY
  { "O_WRONLY"    , VALUE_TYPE_INTEGER  , O_WRONLY    , NULL, NULL },
#endif
#ifdef O_RDWR
  { "O_RDWR"      , VALUE_TYPE_INTEGER  , O_RDWR      , NULL, NULL },
#endif

#ifdef AF_INET
  { "AF_INET"     , VALUE_TYPE_INTEGER  , AF_INET     , NULL, NULL },
#endif
#ifdef SOCK_STREAM
  { "SOCK_STREAM" , VALUE_TYPE_INTEGER  , SOCK_STREAM , NULL, NULL },
#endif
#ifdef SOCK_DGRAM
  { "SOCK_DGRAM"  , VALUE_TYPE_INTEGER  , SOCK_DGRAM  , NULL, NULL },
#endif

#ifdef NLL_FLOATING_POINT
#ifdef USE_MATHLIB
  { "M_E"         , VALUE_TYPE_FLOAT    , 0           , NULL, NULL },
  { "M_PI"        , VALUE_TYPE_FLOAT    , 0           , NULL, NULL },
#endif
#endif

  { NULL          , VALUE_TYPE_UNDEFINED, 0           , NULL, NULL }
};

#ifdef NLL_FLOATING_POINT
double const_variables_floating[] = {
#ifdef USE_MATHLIB
  M_E,
  M_PI,
#endif
  0.0
};
#endif

static struct variable variables[VARIABLE_MAXNUM];

static variable_t head = NULL;
static variable_t pool = NULL;

int variable_check(void)
{
  variable_t variable;
  int n = 0;

  if (head)
    return NLL_ERRCODE_VARIABLE_NOT_EMPTY;

  for (variable = pool; variable && (n < VARIABLE_MAXNUM); variable = variable->next)
    n++;

  if (variable || (n != VARIABLE_MAXNUM))
    return NLL_ERRCODE_VARIABLE_NOT_EMPTY;

  return 0;
}

static int variable_alloc(variable_t *variablep)
{
  variable_t variable;
  value_t value;
  int r;

  if (!pool)
    return NLL_ERRCODE_VARIABLE_BUFFER_OVER;

  if ((r = value_alloc(&value)) < 0)
    return r;

  variable = pool;
  pool = pool->next;

  memset(variable, 0, sizeof(*variable));

  variable->value = value;

  *variablep = variable;

  return 0;
}

static int variable_free(variable_t variable)
{
  variable_t next;

  for (; variable; variable = next) {
    next = variable->next;

    value_free(variable->value);

    variable->next = pool;
    pool = variable;
  }

  return 0;
}

int variable_del(variable_t variable)
{
  variable_t *p, next;

  for (p = &head; *p;) {
    if (*p == variable) {
      if ((--(variable->refcount) == 0) &&
	  (variable->value->type == VALUE_TYPE_UNDEFINED)) {
	next = (*p)->next;
	(*p)->next = NULL;
	variable_free(*p);
	*p = next;
	continue;
      }
    }
    p = &((*p)->next);
  }

  return 0;
}

int variable_clear(void)
{
  variable_t *p, variable, next;

  for (p = &head; *p;) {
    variable = *p;
    if (variable->refcount == 0) {
      next = variable->next;
      variable->next = NULL;
      variable_free(variable);
      *p = next;
      continue;
    }
    p = &(variable->next);
  }

  return 0;
}

int variable_clean(variable_t variable)
{
  int r;
  if ((r = value_clear(variable->value)) < 0)
    return r;
  return 0;
}

int variable_init(void)
{
  variable_t variable;
  int i;

  head = NULL;
  pool = NULL;
  memset(variables, 0, sizeof(variables));

  for (i = 0; i < VARIABLE_MAXNUM; i++) {
    variable = &variables[i];
    variable->next = pool;
    pool = variable;
  }

  return 0;
}

int variable_get_name(char *dname, const char *name)
{
  const char *p;

  for (p = name; *p; p++) {
    if (p == name) {
      if (isalpha(*p) || (*p == '_'))
	;
#ifdef NLL_MULTIBYTE_NAME
      else if (*p & 0x80)
	;
#endif
      else
	break;
    } else {
      if (isalnum(*p) || (*p == '_'))
	;
#ifdef NLL_MULTIBYTE_NAME
      else if (*p & 0x80)
	;
#endif
      else
	break;
    }
    if (p - name > VARIABLE_NAMELEN)
      return NLL_ERRCODE_VARIABLE_TOO_LONG;
    *(dname++) = toupper(*p);
  }

  *dname = '\0';

  return p - name;
}

static int variable_const(variable_t variable)
{
  int r;
  static struct const_variable *p;
#ifdef NLL_FLOATING_POINT
  double *dp = const_variables_floating;
#endif

  for (p = const_variables; p->name; p++) {
    if (!strcmp(variable->name, p->name)) {
      switch (p->type) {
      case VALUE_TYPE_INTEGER:
	if ((r = value_set_integer(variable->value, p->integer)) < 0)
	  return r;
	break;
      case VALUE_TYPE_STRING:
	if ((r = value_set_string(variable->value, p->string, -1)) < 0)
	  return r;
	break;
#ifdef NLL_FLOATING_POINT
      case VALUE_TYPE_FLOAT:
	if ((r = value_set_float(variable->value, *dp)) < 0)
	  return r;
	break;
#endif
      case VALUE_TYPE_POINTER:
	if ((r = value_set_pointer(variable->value, p->pointer, 0)) < 0)
	  return r;
	break;
      default:
	return NLL_ERRCODE_VALUE_INVALID_TYPE;
      }
      if ((r = value_set_const(variable->value)) < 0)
	return r;
    }
#ifdef NLL_FLOATING_POINT
    if (p->type == VALUE_TYPE_FLOAT)
      dp++;
#endif
  }

  return 0;
}

int variable_get(const char *name, variable_t *variablep)
{
  variable_t variable;
  int r;

  for (variable = head; variable; variable = variable->next) {
    if (variable->name[0] && !strcmp(variable->name, name))
      break;
  }

  if (!variable) {
    if ((r = variable_alloc(&variable)) < 0)
      return r;

    strcpy(variable->name, name);

    if ((r = variable_const(variable)) < 0) {
      variable_free(variable);
      return r;
    }

    variable->next = head;
    head = variable;
  }

  variable->refcount++;

  *variablep = variable;

  return 0;
}

int variable_dump(variable_t variable)
{
  if (!variable)
    variable = head;

  printf("\n-- Variable dump --\n");

  for (; variable; variable = variable->next) {
    printf("%s:\tRef:%d\tValue:", variable->name, variable->refcount);
    value_dump(variable->value);
    printf("\n");
  }

  fflush(stdout);

  return 0;
}
